# encoding=utf-8
# @Time    : 17-5-8
# @File    : resource.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import numpy as np
import base64
import copy
from antgo.utils.encode import *
from jinja2 import Environment, FileSystemLoader
import scipy.misc
import sys
import uuid
if sys.version > '3':
  PY3 = True
else:
  PY3 = False
 
  
PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_ENVIRONMENT = Environment(
    autoescape=False,
    loader=FileSystemLoader(os.path.join(PATH, 'templates')),
    trim_blocks=False)


def render_template(template_filename, context):
  return TEMPLATE_ENVIRONMENT.get_template(template_filename).render(context)


def everything_to_html(data, dump_dir):
  assert(len(data) == 1)
  # list all statistics
  everything_statistics = []
  model_deep_analysis = []
  model_sig_diffs = []
  for ant_name, ant_info in data.items():
    # 0.step cpu statistic
    if 'cpu' in ant_info:
      cpu_model = ant_info['cpu']['cpu_model']
      cpu_max_util = ant_info['cpu']['cpu_max_usage']
      cpu_mean_util = ant_info['cpu']['cpu_mean_usage']
      cpu_median_util = ant_info['cpu']['cpu_median_usage']

      cpu_statistic = {'statistic': {'name': 'cpu',
                                     'value': [
                                         {'name': cpu_model,
                                          'value': [['cpu', 'max util', 'mean util','median util'],
                                                    ['-', cpu_max_util, cpu_mean_util,cpu_median_util]],
                                          'type': "TABLE"}]}}
      everything_statistics.append(cpu_statistic)

      # 1.step memory statistic
      mem_max_util = ant_info['cpu']['mem_max_usage']
      mem_mean_util = ant_info['cpu']['mem_mean_usage']
      mem_median_util = ant_info['cpu']['mem_median_usage']

      memory_statistic = {'statistic': {'name': 'cpu memory',
                                        'value': [
                                            {'name': 'memory',
                                             'value': [['memory', 'max util', 'mean util', 'median util'],
                                                       ['-', mem_max_util, mem_mean_util, mem_median_util]],
                                             'type': "TABLE"}]}}
      everything_statistics.append(memory_statistic)
    
    # 1.step gpu statistic
    if 'gpu' in ant_info:
      gpu_model = ant_info['gpu']['gpu_model']
      gpu_max_util = ant_info['gpu']['gpu_max_usage']
      gpu_mean_util = ant_info['gpu']['gpu_mean_usage']
      gpu_median_util = ant_info['gpu']['gpu_median_usage']
      
      value_list = [['gpu', 'max util', 'mean util','median util']]
      for gpu_i in range(len(gpu_max_util)):
        value_list.append(['%d'%gpu_i,
                           '%0.4f'%gpu_max_util[gpu_i],
                           '%0.4f'%gpu_mean_util[gpu_i],
                           '%0.4f'%gpu_median_util[gpu_i]])
      
      gpu_statistic = {'statistic': {'name': 'gpu',
                                     'value': [
                                       {'name': gpu_model,
                                        'value': value_list,
                                        'type': "TABLE"}]}}
      everything_statistics.append(gpu_statistic)
      
      gpu_mem_max_util = ant_info['gpu']['gpu_mem_max_usage']
      gpu_mem_mean_util = ant_info['gpu']['gpu_mem_mean_usage']
      gpu_mem_median_util = ant_info['gpu']['gpu_mem_median_usage']
      
      value_list = [['memory', 'max util', 'mean util', 'median util']]
      for gpu_i in range(len(gpu_mem_max_util)):
        value_list.append(['%d'%gpu_i,
                           '%0.4f'%gpu_mem_max_util[gpu_i],
                           '%0.4f'%gpu_mem_mean_util[gpu_i],
                           '%0.4f'%gpu_mem_median_util[gpu_i]])
      
      gpu_memory_statistic = {'statistic': {'name': 'memory',
                                            'value': [
                                              {'name': 'gpu memory',
                                               'value': value_list,
                                               'type': "TABLE"}]}}
      everything_statistics.append(gpu_memory_statistic)
    
    # 2.step time statistic
    if 'time' in ant_info:
      elapsed_time_per_sample = '-'
      if 'elapsed_time_per_sample' in ant_info['time']:
          elapsed_time_per_sample = ant_info['time']['elapsed_time_per_sample']
      time_statistic = {'statistic': {'name': 'time',
                                      'value': [
                                          {'name': 'time',
                                           'value': [['time', 'total time', 'per sample time'],
                                                     ['-', ant_info['time']['elapsed_time'], elapsed_time_per_sample]],
                                           'type': "TABLE"}]}}
      everything_statistics.append(time_statistic)

    # 3.step model measures
    if 'measure' in ant_info:
      for ms in ant_info['measure']:
        everything_statistics.append(ms)

    # 4.step model analysis
    if 'analysis' in ant_info:
      model_analysis = ant_info['analysis']
      for measure_name, measure_data in model_analysis.items():
        for analysis_title, analysis_data in measure_data.items():
          if type(analysis_data) == list:
            # for group
            for tag, tag_data in analysis_data:
              model_deep_analysis.append({'analysis_name': measure_name,
                                          'analysis_tag': analysis_title+'-'+tag,
                                          'analysis_data': tag_data})
          else:
            # for global
            model_deep_analysis.append({'analysis_name': measure_name,
                                        'analysis_tag': analysis_title,
                                        'analysis_data': analysis_data})
    
    # 5.step model significant difference
    if 'significant_diff' in ant_info:
      sig_diff_dict = ant_info['significant_diff']
      # x in significant diff matrix
      benchmark_model_names = []
      # y in significant diff matrix
      measure_names = []
      # score matrix
      benchmark_model_sig_diff = []

      for measure_name, sig_diff in sig_diff_dict.items():
        # y
        measure_names.append(measure_name)
        if len(benchmark_model_names) == 0:
          # x
          benchmark_model_names = [bn['name'] for bn in sig_diff]
        
        # score (-1, 0, +1) matrix
        benchmark_model_sig_diff.append([bn['score'] for bn in sig_diff])

      benchmark_model_sig_diff = np.array(benchmark_model_sig_diff)
      benchmark_model_sig_diff = benchmark_model_sig_diff.transpose()
      benchmark_model_sig_diff = benchmark_model_sig_diff.tolist()
      model_sig_diffs.append({'score': benchmark_model_sig_diff,
                              'benchmark': benchmark_model_names,
                              'measure': measure_names})
        
  statistic_visualization = _transform_statistic_to_visualization(everything_statistics)
  analysis_visualization = _transform_analysis_to_visualization(model_deep_analysis, dump_dir)
  model_sig_diffs_visualization = _transform_significant_to_visualization(model_sig_diffs)
  
  # 6.step model eye analysys
  eye={'eye_measures': []}
  for ant_name, ant_info in data.items():
    if 'eye' in ant_info:
      if not os.path.exists(os.path.join(dump_dir, 'eye_analysis')):
        os.makedirs(os.path.join(dump_dir, 'eye_analysis'))
      
      for measure_name, eye_data in ant_info['eye'].items():
        eye_measure = {'NAME': measure_name}
        eye_tags = set()
        eye_samples = []
        for sample_index, sample in enumerate(eye_data):
          sample_id = sample['id']
          sample_category = sample['category']
          sample_score = sample['score']
          sample_tag = sample['tag']
          sample_data = sample['data']
          eye_tags.update(sample_tag)
          
          # 1.step warp to eye sample
          eye_sample = {}
          eye_sample['ID'] = sample_id
          eye_sample['TAGS'] = sample_tag
          eye_sample['SCORE'] = sample_score
          # 2.step warp data to eye sample
          if sample['data_type'] == 'IMAGE':
            if type(sample_data) == str:
              eye_sample['DATA'] = 'data:image/png;base64,%s' % sample_data
            else:
              with open(os.path.join(dump_dir, 'eye_analysis', '%s.png' % str(sample_id)), 'wb') as fp:
                fp.write(sample_data)
              eye_sample['DATA'] = './eye_analysis/%s.png' % str(sample_id)

            eye_sample['DATA_TYPE'] = 'IMAGE'
          elif sample['data_type'] == 'STRING':
            eye_sample['DATA'] = sample_data
            eye_sample['DATA_TYPE'] = 'STRING'
            
          # 3.step warp gt to eye sample
          if sample['category_type'] == 'IMAGE':
            if type(sample_category) == str:
              eye_sample['GT'] = 'data:image/png;base64,%s' % sample_data
            else:
              with open(os.path.join(dump_dir, 'eye_analysis', '%s_gt.png' % str(sample_id)), 'wb') as fp:
                fp.write(sample_category)
              eye_sample['GT'] = './eye_analysis/%s_gt.png' % str(sample_id)
            eye_sample['GT_TYPE'] = 'IMAGE'
          elif sample['category_type'] == 'STRING':
            eye_sample['GT'] = sample_category
            eye_sample['GT_TYPE'] = 'STRING'

          eye_samples.append(eye_sample)
        
        eye_tags_list = []
        for tt_i, tt in enumerate(eye_tags):
          eye_tags_list.append({'NAME': tt, 'ID': tt_i})

        eye_measure['eye_tags'] = eye_tags_list
        eye_measure['eye_samples'] = eye_samples
        
        eye['eye_measures'].append(eye_measure)
  
  context = {
    'measures': statistic_visualization,
    'analysis': analysis_visualization,
    'compare': model_sig_diffs_visualization,
    'eye': eye,
  }

  # to resource
  if not os.path.exists(dump_dir):
    os.makedirs(dump_dir)

  with open(os.path.join(dump_dir, 'statistic-report.html'), 'w') as f:
    ss = render_template('statistic-report.html', context)
    if not PY3:
      ss = ss.encode('utf-8')
    f.write(ss)


def _transform_curve_svg_data(data):
  if type(data) != list:
    data = [data]

  reorganized_data = []
  for curve_data in data:
    if type(curve_data) != np.ndarray:
      curve_data = np.array(curve_data)
    
    if curve_data.size == 0:
      reorganized_data.append([])
      continue
      
    x = curve_data[:, 0]
    y = curve_data[:, 1]

    xlist = x.tolist()
    ylist = y.tolist()

    temp = []
    for xv, yv in zip(xlist, ylist):
      temp.append({'x': str(xv), 'y': str(yv)})

    reorganized_data.append(temp)
  return reorganized_data


def _transform_histogram_svg_data(data):
  reorganized_data = []

  if type(data[0]) != list:
    data = [data]
  for his_data in data:
    svg_data = []
    for xv, yv in zip(range(0, len(his_data)), his_data):
      svg_data.append({'x': str(xv), 'y': str(yv)})

    reorganized_data.append(svg_data)

  return reorganized_data


def _transform_image_data(data):
  ss = base64.b64encode(png_encode(data))
  ss = ss.decode('utf-8')
  return ss


def _transform_statistic_to_visualization(statistic_info):
  visualization_statistic_info = []
  for sta_info in statistic_info:
    for index in range(len(sta_info['statistic']['value'])):
      if sta_info['statistic']['value'][index]['type'] == 'CURVE':
        statistic_value = sta_info['statistic']['value'][index]['value']
        reorganized_statistic_value = _transform_curve_svg_data(statistic_value)

        sta_info_cpy = copy.deepcopy(sta_info)
        sta_info_cpy['statistic']['value'] = [sta_info['statistic']['value'][index]]
        sta_info_cpy['statistic']['value'][0]['value'] = reorganized_statistic_value
        visualization_statistic_info.append(sta_info_cpy)
      elif sta_info['statistic']['value'][index]['type'] == 'SCALAR':
        statistic_value = sta_info['statistic']['value'][index]['value']
        if type(statistic_value) == list:
          # histgram
          reorganized_statistic_value = _transform_histogram_svg_data(statistic_value)

          sta_info_cpy = copy.deepcopy(sta_info)
          sta_info_cpy['statistic']['value'] = [sta_info['statistic']['value'][index]]
          sta_info_cpy['statistic']['value'][0]['value'] = reorganized_statistic_value
          sta_info_cpy['statistic']['value'][0]['type'] = 'HISTOGRAM'
          visualization_statistic_info.append(sta_info_cpy)
        else:
          # scalar
          sta_info_cpy = copy.deepcopy(sta_info)
          sta_info_cpy['statistic']['value'] = [sta_info['statistic']['value'][index]]
          sta_info_cpy['statistic']['value'][0]['value'] = str(sta_info_cpy['statistic']['value'][0]['value'])
          if 'interval' in sta_info_cpy['statistic']['value'][0]:
              sta_info_cpy['statistic']['value'][0]['interval'] = str(sta_info_cpy['statistic']['value'][0]['interval'])
          visualization_statistic_info.append(sta_info_cpy)
      elif sta_info['statistic']['value'][index]['type'] == 'IMAGE':
        image_value = sta_info['statistic']['value'][index]['value']
        # image
        # reorganized_statistic_value = _transform_image_data(image_value)
        reorganized_statistic_value = image_value

        sta_info_cpy = copy.deepcopy(sta_info)
        sta_info_cpy['statistic']['value'] = [sta_info['statistic']['value'][index]]
        sta_info_cpy['statistic']['value'][0]['value'] = reorganized_statistic_value
        sta_info_cpy['statistic']['value'][0]['type'] = 'IMAGE'
        visualization_statistic_info.append(sta_info_cpy)

      else:
        # TABLE / MATRIX
        sta_info_cpy = copy.deepcopy(sta_info)
        sta_info_cpy['statistic']['value'] = [sta_info['statistic']['value'][index]]
        sta_info_cpy['statistic']['value'][0]['value'] = str(sta_info_cpy['statistic']['value'][0]['value'])
        visualization_statistic_info.append(sta_info_cpy)

  return visualization_statistic_info


def _transform_analysis_to_visualization(analysis_info, dump_dir):
  analysis_vis_data = []
  # analysis-name, analysis-tag, analysis_data
  for data in analysis_info:
    analysis_name = data['analysis_name']
    analysis_tag = data['analysis_tag']
    analysis_data = data['analysis_data']
    item = {}
    item['name'] = '%s-%s'%(analysis_name, analysis_tag)
    item['value'] = analysis_data['value'].tolist() if type(analysis_data['value']) != list else analysis_data['value']
    if type(analysis_data['x']) == list:
      item['x'] = analysis_data['x']
    else:
      item['x'] = analysis_data['x'].tolist()
    
    if type(analysis_data['y']) == list:
      item['y'] = analysis_data['y']
    else:
      item['y'] = analysis_data['y'].tolist()

    region_samplings = analysis_data['sampling']
    region_samplings_vis = []
    for sampling in region_samplings:
      mmm = []
      for sampling_data in sampling['data']:
        if sampling_data['type'] == 'IMAGE':
          if type(sampling_data['data']) == str:
            mmm.append({'data':'data:image/png;base64,%s'%sampling_data['data'], 'flag': '', 'type': sampling_data['type']})
          else:
            if not os.path.exists(os.path.join(dump_dir, 'statistic_analysis')):
              os.makedirs(os.path.join(dump_dir, 'statistic_analysis'))

            uuid_name = '%s.png' % str(uuid.uuid4())
            with open(os.path.join(dump_dir, 'statistic_analysis', uuid_name), 'wb') as fp:
              fp.write(sampling_data['data'])

            mmm.append(
              {'data': './statistic_analysis/%s'%uuid_name, 'flag': 'image_container', 'type': sampling_data['type']})
        else:
          mmm.append({'data': sampling_data['data'], 'type': sampling_data['type']})

      region_samplings_vis.append({'name': sampling['name'], 'data': mmm})

    item['region_samplings'] = region_samplings_vis
    analysis_vis_data.append(item)

  return analysis_vis_data


def _transform_significant_to_visualization(sig_diffs):
  return sig_diffs

if __name__ == '__main__':
  # experiment 1
  experiment_1_statis = {}
  experiment_1_statis['aa'] = {}
  experiment_1_statis['aa']['time'] = {}
  experiment_1_statis['aa']['time']['elapsed_time'] = 1.2
  experiment_1_statis['aa']['cpu'] = {}
  experiment_1_statis['aa']['cpu']['mem_mean_usage'] = 23
  experiment_1_statis['aa']['cpu']['mem_median_usage'] = 12
  experiment_1_statis['aa']['cpu']['mem_max_usage'] = 44
  experiment_1_statis['aa']['cpu']['cpu_mean_usage'] = 55
  experiment_1_statis['aa']['cpu']['cpu_median_usage'] = 11
  experiment_1_statis['aa']['cpu']['cpu_max_usage'] = 22
  experiment_1_statis['aa']['cpu']['cpu_model'] = 'aabbcc'

  voc_measure = {'statistic': {'name': 'voc',
                               'value': [{'name': 'MAP', 'value': [23.0, 11.0, 12.0], 'type': 'SCALAR', 'x': 'class',
                                          'y': 'Mean Average Precision'},
                                         {'name': 'Mean-MAP', 'value': 0.23, 'type': 'SCALAR'}]}}

  roc_auc_measure = {'statistic': {'name': 'roc_auc',
                                   'value': [{'name': 'ROC', 'value': [[[0,0],[1,1],[2,3]],
                                                                       [[0,3],[2,5],[3,0]]],
                                              'type': 'CURVE', 'x': 'FP', 'y': 'TP', 'legend':['class-0','class-1']},
                                             {'name': 'AUC', 'value': [0.3, 0.4], 'type': 'SCALAR', 'x': 'class',
                                              'y': 'AUC'}]}}

  pr_f1_measure = {'statistic': {'name': 'pr_f1',
                                 'value': [{'name': 'Precision-Recall',
                                            'value': [[[0,0],[1,1],[2,3]],
                                                                       [[0,3],[2,5],[3,0]]],
                                            'type': 'CURVE', 'x': 'precision', 'y': 'recall'},
                                           {'name': 'F1', 'value': [1.0, 2.0], 'type': 'SCALAR', 'x': 'class',
                                            'y': 'F1'}]}}

  confusion_m = {'statistic': {'name': 'cm',
                               'value': [{'name': 'ccmm', 'value': (np.ones((3, 4)) * 3).tolist(), 'type': 'MATRIX',
                                          'x': ['a','b','c','d'], 'y': ['x','y','z']}]}}

  random_img = np.random.random((100,100))
  random_img = random_img * 255
  random_img = random_img.astype(np.uint8)
  image_m = {'statistic': {'name': 'image',
                               'value': [{'name': 'image', 'value': random_img, 'type': 'IMAGE'}]}}

  experiment_1_statis['aa']['measure'] = [voc_measure, roc_auc_measure, pr_f1_measure, confusion_m, image_m]
  experiment_1_statis['aa']['eye'] = {}
  experiment_1_statis['aa']['eye']['MM'] = [{'id': 0, 'category': 'hello', 'score':0.2, 'tag':['A','B'], 'data': 'mma'},
                                            {'id': 3, 'category': 'miao', 'score': 0.4, 'tag':['C'], 'data': 'yYU'},
                                            {'id': 1, 'category': 'world', 'score':1.0, 'tag':['A'], 'data': 'UUI'}]


  # analysis data
  scores = np.random.randint(0,100,200)
  scores = scores.reshape((2,100))

  sample_index = list(range(100))
  image = cv2.imread('/Users/jian/Downloads/owl1.jpg')
  image = image[...,::-1]
  # image = base64.b64encode(png_encode(image))
  # image = image.decode('utf-8')
  hr_samples = []
  mr_samples = []
  lr_samples = []
  for _ in range(10):
    hr_samples.append({'type': 'IMAGE', 'data': image})
    mr_samples.append({'type': 'IMAGE', 'data': image})
    lr_samples.append({'type': 'IMAGE', 'data': image})

  experiment_1_statis['aa']['analysis'] = {}
  experiment_1_statis['aa']['analysis']['AA']={}
  experiment_1_statis['aa']['analysis']['AA']['Global']={
    'value': scores,
    'type': 'MATRIX',
    'x': sample_index,
    'y': ['12345.1231.134134521','20142312.1324.135234'],
    'sampling': [{'name': 'High Score Region', 'data': hr_samples},
                 {'name': 'Middle Score Region', 'data': mr_samples},
                 {'name': 'Low Score Region', 'data': lr_samples}],
  }

  #
  # import cv2
  mm = cv2.imread('/Users/jian/Downloads/owl1.jpg')
  mm = mm[...,::-1]

  experiment_1_statis['aa']['eye']['NN'] = [{'id': 0, 'category': mm, 'score':0.2, 'tag':['A','B'], 'data': mm},
                                            {'id': 3, 'category': mm, 'score': 0.4, 'tag':['C'], 'data': mm},
                                            {'id': 1, 'category': mm, 'score':1.0, 'tag':['A'], 'data': mm}]
  
  # # experiment 2
  # experiment_2_statis = {}
  # experiment_2_statis['aa'] = {}
  # experiment_2_statis['aa']['time'] = {}
  # experiment_2_statis['aa']['time']['elapsed_time'] = 1.2
  # experiment_2_statis['aa']['cpu'] = {}
  # experiment_2_statis['aa']['cpu']['mem_mean_usage'] = 23
  # experiment_2_statis['aa']['cpu']['mem_median_usage'] = 12
  # experiment_2_statis['aa']['cpu']['mem_max_usage'] = 44
  # experiment_2_statis['aa']['cpu']['cpu_mean_usage'] = 55
  # experiment_2_statis['aa']['cpu']['cpu_median_usage'] = 11
  # experiment_2_statis['aa']['cpu']['cpu_max_usage'] = 22
  # experiment_2_statis['aa']['cpu']['cpu_model'] = 'aabbcc'
  #
  # voc_measure = {'statistic': {'name': 'voc',
  #                              'value': [{'name': 'MAP', 'value': [2.0, 1.0, 2.0], 'type': 'SCALAR', 'x': 'class',
  #                                         'y': 'Mean Average Precision'},
  #                                        {'name': 'Mean-MAP', 'value': 0.12, 'type': 'SCALAR'}]}}
  #
  # roc_auc_measure = {'statistic': {'name': 'roc_auc',
  #                                  'value': [
  #                                      {'name': 'ROC',
  #                                       'value': [(np.ones((3, 2)) * 4).tolist(), (np.ones((3, 2)) * 1).tolist()],
  #                                       'type': 'CURVE',
  #                                       'x': 'FP', 'y': 'TP'},
  #                                      {'name': 'AUC', 'value': [0.12, 0.34], 'type': 'SCALAR', 'x': 'class',
  #                                       'y': 'AUC'}]}}
  #
  # pr_f1_measure = {'statistic': {'name': 'pr_f1',
  #                                'value': [
  #                                    {'name': 'Precision-Recall',
  #                                     'value': [(np.ones((4, 2)) * 1).tolist(), (np.ones((4, 2)) * 2).tolist()],
  #                                     'type': 'CURVE', 'x': 'precision', 'y': 'recall'},
  #                                    {'name': 'F1', 'value': [1.4, 1.0], 'type': 'SCALAR', 'x': 'class', 'y': 'F1'}]}}
  #
  # confusion_m = {'statistic': {'name': 'cm',
  #                              'value': [{'name': 'ccmm', 'value': (np.ones((3, 4)) * 8).tolist(), 'type': 'MATRIX',
  #                                         'x': 'class', 'y': 'class'}]}}
  #
  # experiment_2_statis['aa']['measure'] = [voc_measure, roc_auc_measure, pr_f1_measure, confusion_m]

  # ss = multi_repeats_measures_statistic([experiment_1_statis, experiment_2_statis])
  everything_to_html(experiment_1_statis, '/Users/jian/Downloads/')


# -*- coding: UTF-8 -*-
# @Time    : 17-12-28
# @File    : image_tool.py
# @Author  : 
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import matplotlib.pyplot as plt
import matplotlib.patches as patches

def imshow(image, bboxes=None, keypoints=None):
  # Create figure and axes
  fig, ax = plt.subplots(1)
  
  # Display the image
  image = image.copy()
  if keypoints is not None:
    for k in range(keypoints.shape[0]):
      x, y = keypoints[k]
      image[int(y), int(x), :] = 255
      
  ax.imshow(image)
  
  if bboxes is not None:
    for b in range(bboxes.shape[0]):
      x0, y0, x1, y1 = bboxes[b]
      width = x1 - x0
      height = y1 - y0
      # Create a Rectangle patch
      rect = patches.Rectangle((x0, y0), width, height, linewidth=1, edgecolor='r', facecolor='none')
      # Add the patch to the Axes
      ax.add_patch(rect)

  plt.show()
  plt.waitforbuttonpress()


# encoding=utf-8
# @Time    : 17-6-8
# @File    : grid.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.core import *
from antgo.codebook.bboxes.bboxes import *
import copy


class Grid(Node):
  def __init__(self, inputs, grid=(), cell_size=0.5, obj_overlap=0.5):
    super(Grid, self).__init__(name=None, action=self.action, inputs=inputs)
    self._grid = grid
    self._cell_size = cell_size
    self._obj_overlap = obj_overlap

  def _grid_crop(self, image_size, grid, cell_ratio):
    width, height = image_size[1::-1]
    square_cell_size = np.minimum(width * cell_ratio, height * cell_ratio)
    grid_hori_num = grid[0]
    grid_vert_num = grid[1]

    hori_step = 0
    if grid_hori_num >= 2:
      hori_step = int(float(width - square_cell_size) / float(grid_hori_num - 1))
    vert_step = 0
    if grid_vert_num >= 2:
      vert_step = int(float(height - square_cell_size) / float(grid_vert_num - 1))

    crops = np.zeros((grid_hori_num * grid_vert_num, 4), np.int32)
    for grid_hori_index in range(grid_hori_num):
      for grid_vert_index in range(grid_vert_num):
        crop_index = grid_hori_index * grid_vert_num + grid_vert_index
        crops[crop_index, 0] = hori_step * grid_hori_index
        crops[crop_index, 1] = vert_step * grid_vert_index

        crops[crop_index, 2] = crops[crop_index, 0] + square_cell_size
        crops[crop_index, 3] = crops[crop_index, 1] + square_cell_size

    return crops

  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    data, annotation = args[0] if len(args[0]) == 2 else (args[0], {})
    assert(type(data) == np.ndarray)
    crops = self._grid_crop(data.shape, self._grid, self._cell_size)

    patches = []
    for xy in np.split(crops, crops.shape[0], axis=0):
      patch_x0, patch_y0, patch_x1, patch_y1 = xy[0, :]
      # crop
      patch = data[int(patch_y0):int(patch_y1), int(patch_x0):int(patch_x1)]

      annotation_cpy = copy.deepcopy(annotation)
      if 'bbox' in annotation_cpy:
        bboxes = annotation_cpy['bbox']
        remained_bboxes, remained_bboxes_ind = \
          bboxes_filter_overlap(np.array((patch_x0,patch_y0,patch_x1,patch_y1)), bboxes, self._obj_overlap)

        remained_bboxes = bboxes_translate(np.array((patch_x0,patch_y0,patch_x1,patch_y1)), remained_bboxes)
        annotation_cpy['bbox'] = remained_bboxes

        if 'category_id' in annotation_cpy:
          annotation_cpy['category_id'] = annotation_cpy['category_id'][remained_bboxes_ind]
        if 'category' in annotation_cpy:
          annotation_cpy['category'] = [annotation_cpy['category'][i] for i in remained_bboxes_ind]
        if 'area' in annotation_cpy:
          annotation_cpy['area'] = annotation_cpy['area'][remained_bboxes_ind]

        if 'segmentation' in annotation_cpy:
          annotation_cpy['segmentation'] = \
            [annotation_cpy['segmentation'][i][int(patch_y0):int(patch_y1), int(patch_x0):int(patch_x1)]
             for i in remained_bboxes_ind]

      annotation_cpy['cell'] = xy[0, :]
      annotation_cpy['info'] = patch.shape
      annotation_cpy['image'] = data
      patches.append((patch, annotation_cpy))
    return patches


def sample_grid_integrate(cells, bboxes):
  '''
  :param cells: [[x0,y0,x1,y1],[],[],...]
  :param bboxes: grid_cells x bboxes_num x 4
  :return:
  '''
  # translate all bboxes in every grid cell
  cells_num = cells.shape[0]
  for cell_index in range(cells_num):
    cell_offset_x = cells[cell_index,0]
    cell_offset_y = cells[cell_index,1]
    bboxes[cell_index, :, :] = bboxes[cell_index, :, :] + \
                               np.array((cell_offset_x, cell_offset_y, cell_offset_x, cell_offset_y))

  return bboxes.reshape(-1, 4)


# encoding=utf-8
# @Time    : 17-5-8
# @File    : statistic.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from contextlib import contextmanager
from collections import defaultdict
from ..utils.cpu import *
from ..utils.gpu import *
from ..utils.concurrency import *
from antgo.utils import logger
import time
import psutil
import os
import numpy as np
import subprocess

_RUNNING_STATISTIC = defaultdict(dict)
_RUNNING_STATISTIC_TREE = []

@contextmanager
def performance_statistic_region(running_block):
  global _RUNNING_STATISTIC
  if running_block in _RUNNING_STATISTIC:
    _RUNNING_STATISTIC.pop(running_block)

  _RUNNING_STATISTIC_TREE.append(running_block)
  statistic_tree_block = '/'.join(_RUNNING_STATISTIC_TREE)

  # current process
  pid = os.getpid()

  # now cpu status
  now_cpu_status = cpu_running_info(pid)
  running_mem_usage = [now_cpu_status['cpu_mem_usage']]
  running_cpu_util = [now_cpu_status['cpu_util']]
  running_cpu_occupy = [now_cpu_status['occupy_cpus']]
  cpu_model = [now_cpu_status['cpus']]

  def _cpu_info_func(pid):
    try:
      cpu_info = cpu_running_info(pid)
      mem_usage = cpu_info['cpu_mem_usage']
      cpu_util = cpu_info['cpu_util']
      cpu_occupy = cpu_info['occupy_cpus']
      running_mem_usage.append(mem_usage)
      running_cpu_util.append(cpu_util)
      running_cpu_occupy.append(cpu_occupy)
    except:
      logger.error('some error happen when sampling cpu state')

  # now gpu status
  now_gpu_status = gpu_running_info(pid)
  gpu_model = [now_gpu_status['gpus']] if now_gpu_status is not None else []
  gpu_driver_version = [now_gpu_status['driver-version']] if now_gpu_status is not None else []
  
  running_gpu_mem_usage = [[float(now_gpu_status['gpu_mem_usage'][int(index)]) for index in now_gpu_status['occupy_gpus']]] \
                          if now_gpu_status is not None else []
  running_gpu_util = [[float(now_gpu_status['gpu_util'][int(index)]) for index in now_gpu_status['occupy_gpus']]] \
                      if now_gpu_status is not None else []
  running_gpu_occupy = [now_gpu_status['occupy_gpus']] if now_gpu_status is not None else []

  def _gpu_info_func(pid):
    try:
      gpu_info = gpu_running_info(pid)
      if gpu_info is not None:
        assert(len(gpu_info['occupy_gpus']) <= len(gpu_info['gpu_mem_usage']))
        assert(len(gpu_info['occupy_gpus']) <= len(gpu_info['gpu_util']))
        assert(len(gpu_info['gpu_mem_usage']) == len(gpu_info['gpu_util']))
        if len(gpu_info['occupy_gpus']) == 0:
          return
        
        running_gpu_mem_usage.append([float(gpu_info['gpu_mem_usage'][int(index)]) for index in gpu_info['occupy_gpus']])
        running_gpu_util.append([float(gpu_info['gpu_util'][int(index)]) for index in gpu_info['occupy_gpus']])
        running_gpu_occupy.append(gpu_info['occupy_gpus'])
  
        if len(gpu_model) == 0:
          gpu_model.append(gpu_info['gpus'])
        
        if len(gpu_driver_version) == 0:
          gpu_driver_version.append(gpu_info['driver-version'])
    except:
      logger.error('some error happen when sampling gpu state')

  # running before
  # 1.step running time start
  start = time.time()

  # 2.step cpu and gpu statistic periodically
  timer_thread = TimerThread([lambda:_cpu_info_func(pid), lambda:_gpu_info_func(pid)], periodic=5)
  timer_thread.start()

  yield
  
  # stop thread
  timer_thread.stop()

  # running after
  elapsed_time = time.time() - start

  time_length = len(running_mem_usage)
  sta_tstart = int(time_length * 0.3)
  sta_tstop = int(time_length * 0.8)

  if sta_tstart == sta_tstop:
    sta_tstop = sta_tstart + 1
  
  # memory util status
  running_mem_usage_mean = np.mean(running_mem_usage[sta_tstart:sta_tstop])
  running_mem_usage_median = np.median(running_mem_usage[sta_tstart:sta_tstop])
  running_mem_usage_max = np.max(running_mem_usage[sta_tstart:sta_tstop])

  # cpu util status
  running_cpu_util_mean = np.mean(running_cpu_util[sta_tstart:sta_tstop])
  running_cpu_util_median = np.median(running_cpu_util[sta_tstart:sta_tstop])
  running_cpu_util_max = np.max(running_cpu_util[sta_tstart:sta_tstop])
  _RUNNING_STATISTIC[statistic_tree_block]['time'] = {}
  _RUNNING_STATISTIC[statistic_tree_block]['time']['elapsed_time'] = elapsed_time
  _RUNNING_STATISTIC[statistic_tree_block]['cpu'] = {}
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['mem_mean_usage'] = running_mem_usage_mean
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['mem_median_usage'] = running_mem_usage_median
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['mem_max_usage'] = running_mem_usage_max
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['cpu_mean_usage'] = running_cpu_util_mean
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['cpu_median_usage'] = running_cpu_util_median
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['cpu_max_usage'] = running_cpu_util_max
  _RUNNING_STATISTIC[statistic_tree_block]['cpu']['cpu_model'] = cpu_model[0][0]

  if len(gpu_model) > 0:
    # sampling gpu
    time_length = len(running_gpu_occupy)
    sta_tstart = int(time_length * 0.3)
    sta_tstop = int(time_length * 0.8)
  
    if sta_tstart == sta_tstop:
      sta_tstop = sta_tstart + 1
    
    # filter invalid gpu sampling
    occupied_gpu_num = 0
    for sampling_i in range(sta_tstart, sta_tstop, 1):
      if occupied_gpu_num < len(running_gpu_occupy[sampling_i]):
        occupied_gpu_num = len(running_gpu_occupy[sampling_i])
    
    if occupied_gpu_num > 0:
      running_gpu_mem_usage = [vv for vv in running_gpu_mem_usage if len(vv) == occupied_gpu_num]
      running_gpu_util = [vv for vv in running_gpu_util if len(vv) == occupied_gpu_num]
      
      # gpu memory util status
      running_gpu_mem_usage_mean = np.mean(np.array(running_gpu_mem_usage[sta_tstart:sta_tstop]), axis=0).tolist()
      running_gpu_mem_usage_median = np.median(running_gpu_mem_usage[sta_tstart:sta_tstop], axis=0).tolist()
      running_gpu_mem_usage_max = np.max(running_gpu_mem_usage[sta_tstart:sta_tstop], axis=0).tolist()
  
      # gpu util status
      running_gpu_util_usage_mean = np.mean(np.array(running_gpu_util[sta_tstart:sta_tstop]), axis=0).tolist()
      running_gpu_util_usage_median = np.median(running_gpu_util[sta_tstart:sta_tstop], axis=0).tolist()
      running_gpu_util_usage_max = np.max(running_gpu_util[sta_tstart:sta_tstop], axis=0).tolist()
  
      _RUNNING_STATISTIC[statistic_tree_block]['gpu'] = {}
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_mem_mean_usage'] = running_gpu_mem_usage_mean
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_mem_median_usage'] = running_gpu_mem_usage_median
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_mem_max_usage'] = running_gpu_mem_usage_max
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_mean_usage'] = running_gpu_util_usage_mean
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_median_usage'] = running_gpu_util_usage_median
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_max_usage'] = running_gpu_util_usage_max
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_model'] = gpu_model[0][0]
      _RUNNING_STATISTIC[statistic_tree_block]['gpu']['gpu_driver_version'] = gpu_driver_version[0]

  # waiting until stop
  timer_thread.join()
  _RUNNING_STATISTIC_TREE.pop(-1)


def get_performance_statistic(running_block=None):
  return _RUNNING_STATISTIC if running_block is None else _RUNNING_STATISTIC[running_block]



# -*- coding: UTF-8 -*-
# @Time    : 2020-07-04 10:24
# @File    : batch_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import sys
import uuid
import json
import time
import math


class EntryApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    content = self.db['content']
    if content is None:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if 'dataset' not in content:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    tags = content['tags']
    dataset_and_results = content['dataset']
    dataset_names = list(dataset_and_results.keys())
    dataset_name = dataset_names[0]

    query_dataset_name = self.get_argument('dataset_name', None)
    if query_dataset_name is not None:
      dataset_name = query_dataset_name
    result_num = len(dataset_and_results[dataset_name])
    current_dataset_name = dataset_name

    if dataset_name not in self.db:
      self.db[dataset_name] = {}

    # 每页只能显示100条记录(固定)
    num_in_page = 100
    # 获得需要展示的页数
    page_num = int(math.ceil(result_num / num_in_page))
    self.db[dataset_name]['num_in_page'] = num_in_page
    self.db[dataset_name]['page_num'] = page_num
    page_index = 0
    start_index = 0 * num_in_page
    start_index = start_index if start_index < result_num else result_num-1
    end_index = 1 * num_in_page
    end_inidex = end_index if end_index < result_num else result_num

    sample_num_in_dataset = {}
    for dataset_name in dataset_names:
      sample_num_in_dataset[dataset_name] = len(self.db['content']['dataset'][dataset_name])

    response_content = {
      'num_in_page': num_in_page,
      'page_num': page_num,
      'page_index': page_index,
      'dataset_name': dataset_names,
      'sample_num_in_dataset': sample_num_in_dataset,
      'current_dataset_name': current_dataset_name,
      'result_num': result_num,
      'result_list': dataset_and_results[dataset_name][start_index: end_index],
      'tags': tags,
      'waiting': self.db['waiting'],
      'finished': self.db['finished'],
      'spider': self.db['spider'] if 'spider' in self.db else 'BAIDU'
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)

class TagChangeApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    page_index = self.get_argument('page_index', None)
    data_index = self.get_argument('data_index', None)
    data_tag = self.get_argument('data_tag', None)
    if page_index is None or data_index is None or data_tag is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
    
    page_index = (int)(page_index)
    data_index = (int)(data_index)
    data_tag = json.loads(data_tag)

    # 修改数据的tag
    dataset_and_results = self.db['content']['dataset']
    dataset_names = list(dataset_and_results.keys())
    dataset_name = dataset_names[0]
    num_in_page = self.db[dataset_name]['num_in_page']

    data_index = page_index * num_in_page + data_index
    self.db['content']['dataset'][dataset_name][data_index]['tag'] = data_tag
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class SearchApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    web_url = self.get_argument('url', None)
    keyword = self.get_argument('keyword', None)
    if web_url.lower().startswith('baidu'):
      web_url = 'baidu'
    elif web_url.lower().startswith('bing'):
      web_url = 'bing'
    elif web_url.lower().startswith('google'):
      web_url = 'google'

    # 记录当前spider
    self.db['spider'] = web_url.upper()

    # 通知爬虫
    self.db['command_queue'].put((
      'baidu', {'datasource_address': web_url, 'datasource_keywards': keyword}
    ))
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class PageApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    page_index = self.get_argument('page_index', None)
    page_index = int(page_index)
    dataet_name = self.get_argument('dataset_name', None)
    if page_index is None or dataet_name is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    content = self.db['content']
    if content is None:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if dataet_name not in self.db:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if 'num_in_page' not in self.db[dataet_name] or 'page_num' not in self.db[dataet_name]:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if page_index >= self.db[dataet_name]['page_num']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    result_num = len(self.db['content']['dataset'][dataet_name])
    start_index = page_index * self.db[dataet_name]['num_in_page']
    start_index = start_index if start_index < result_num else result_num - 1
    end_index = (page_index + 1) * self.db[dataet_name]['num_in_page']
    end_inidex = end_index if end_index < result_num else result_num

    dataset_names = list(content['dataset'].keys())
    sample_num_in_dataset = {}
    for dataset_name in dataset_names:
      sample_num_in_dataset[dataset_name] = len(self.db['content']['dataset'][dataset_name])

    response_content = {
      'num_in_page': self.db[dataet_name]['num_in_page'],
      'page_num': self.db[dataet_name]['page_num'],
      'page_index': page_index,
      'dataset_name': dataset_names,
      'current_dataset_name': dataet_name,
      'sample_num_in_dataset': sample_num_in_dataset,
      'result_num': result_num,
      'result_list': content['dataset'][dataet_name][start_index: end_index],
      'waiting': self.db['waiting'],
      'finished': self.db['finished']
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class ConfigApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    config_data = self.get_argument('config_data', None)
    if config_data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    # config_data: 
    # {'experiment_uuid': experiment_uuid, 'dataset': {'train': [[{},{},...],[]]}, 'tags':[]}
    config_data = json.loads(config_data)
    if len(self.db['content']) == 0:
      self.db['content'] = config_data
    else:
      for key in config_data['dataset'].keys():
        self.db['content']['dataset'][key].extend(config_data['dataset'][key])

    self.db['waiting'] = config_data['waiting']
    self.db['finished'] = config_data['finished']
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class PingApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def batch_server_start(batch_dump_dir, server_port, command_queue):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    batch_static_dir = os.path.join(batch_dump_dir, 'batch')

    # 2.step launch web server
    db = {'content': {}, 'command_queue': command_queue}
    settings = {
      'static_path': os.path.join(batch_static_dir, 'static'),
      'port': server_port,
      'cookie_secret': str(uuid.uuid4()),
      'db': db,
    }

    app = tornado.web.Application(handlers=[(r"/batch-api/entry/", EntryApiHandler),
                                            (r"/batch-api/dataset/", EntryApiHandler),
                                            (r"/batch-api/page/", PageApiHandler),
                                            (r"/batch-api/config/", ConfigApiHandler),
                                            (r"/batch-api/ping/", PingApiHandler),
                                            (r"/batch-api/tag/", TagChangeApiHandler),
                                            (r"/batch-api/search/", SearchApiHandler),
                                            (r'/(.*)', tornado.web.StaticFileHandler,
                                             {"path": batch_static_dir, "default_filename": "index.html"}),],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('demo is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass


# encoding=utf-8
# @Time    : 17-5-4
# @File    : regression_task.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
from antgo.measures.regression_metric import *


default = {'AntMAPERegression': ('MAPE', 'REGRESSION'),
           'AntAlmostCRegression': ('ALMOST-CORRECT', 'REGRESSION')}
class AntMAPERegression(AntMeasure):
  def __init__(self, task):
    super(AntMAPERegression, self).__init__(task, 'MAPE')
    assert(task.task_type == 'REGRESSION')
    self.is_support_rank = True

  def eva(self, data, label):
    '''
    :param data: predicate value (N,)
    :param label: ground truth value (N,)
    :return:
    '''
    # assert(data.shape[0] == label.shape[0])
    if label is not None:
      data = zip(data, label)

    acutal_s = []
    predicated_s = []
    sample_scores = []
    for predict, gt in data:
      predicated_s.append(predict)

      id = None
      gt_label = gt
      if type(gt) == dict:
        gt_label = float(gt['data'])
        id = gt['id']

      if id is not None:
        sample_scores.append({'id': id, 'score': abs(gt_label - predict), 'category': gt_label})
      acutal_s.append(gt_label)

    error = mape(actual_s=acutal_s, predicated_s=predicated_s)
    return {'statistic': {'name': self.name,
                          'value': [{'name':self.name, 'value': error, 'type': 'SCALAR'}]},
            'info': sample_scores}


class AntAlmostCRegression(AntMeasure):
  def __init__(self, task):
    super(AntAlmostCRegression, self).__init__(task,'ALMOST-CORRECT')
    assert(task.task_type == 'REGRESSION')
    self.is_support_rank = True

  def eva(self, data, label):
    '''
    :param data: predicate value (N,)
    :param label: ground truth value (N,)
    :return:
    '''
    # assert(data.shape[0] == label.shape[0])
    if label is not None:
      data = zip(data, label)

    almost_degree = int(getattr(self.task, 'almost_correct', 1.0))

    acutal_s = []
    predicated_s = []
    sample_scores = []
    for predict, gt in data:
      predicated_s.append(predict)

      id = None
      gt_label = gt
      if type(gt) == dict:
        gt_label = float(gt['data'])
        id = gt['id']

      if id is not None:
        sample_scores.append({'id': id,
                              'score': 1 if abs(gt_label - predict) < almost_degree else 0,
                              'category': gt_label})

      acutal_s.append(gt)

    error = almost_correct(acutal_s, predicated_s, almost_degree)
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': error, 'type': 'SCALAR'}]},
            'info': sample_scores}